/*
(C) 2013 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//This code uses the un-divided 8MHz internal oscillator. Please adjust your fuses accordingly.
#define F_CPU 8000000L

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdint.h>


#define MOT1A PB0
#define MOT1B PB1
#define MOT1C PB2
#define MOT1D PB3
#define MOT2A PB4
#define MOT2B PB5
#define MOT2C PB6
#define MOT2D PB7
#define BTN PD1
#define SPIT PD2
#define MOT3A PD3
#define MOT3B PD4
#define MOT3C PD5
#define MOT3D PD6


#define MOTA 1
#define MOTB 2
#define MOTC 4
#define MOTD 8
volatile const char motorStates[] PROGMEM={
	MOTA,
	MOTA|MOTD,
	     MOTD,
	MOTB|MOTD,
	MOTB,
	MOTB|MOTC,
	     MOTC,
    MOTA|MOTC
};



volatile int motval[3]={0,0,0}; //Current motor values
volatile char spit=0;			//Does the syringe eject food colouring?
volatile char disable=1;		//If true, turn off motors to save power

//Set all outputs to 0 so no current can run through the motors.
void disableMotors() {
	PORTB=0;
	PORTD&=~(0x78);
}

//This sets a motor to the position given in the first 3 bits of val.
//It'll look up the corresponding motor line states in motorStates and send
//the value to the specified port.
void setMotor(int motor, int val) {
	char c=pgm_read_byte(&motorStates[val&7]);
	if (disable) {
		disableMotors();
		PORTD&=~(1<<SPIT);
		return;
	}
	if (motor==0) {
		if (c&MOTA) PORTB|=(1<<MOT1A); else PORTB&=~(1<<MOT1A);
		if (c&MOTB) PORTB|=(1<<MOT1B); else PORTB&=~(1<<MOT1B);
		if (c&MOTC) PORTB|=(1<<MOT1C); else PORTB&=~(1<<MOT1C);
		if (c&MOTD) PORTB|=(1<<MOT1D); else PORTB&=~(1<<MOT1D);
	} else if (motor==1) {
		if (c&MOTA) PORTB|=(1<<MOT2A); else PORTB&=~(1<<MOT2A);
		if (c&MOTB) PORTB|=(1<<MOT2B); else PORTB&=~(1<<MOT2B);
		if (c&MOTC) PORTB|=(1<<MOT2C); else PORTB&=~(1<<MOT2C);
		if (c&MOTD) PORTB|=(1<<MOT2D); else PORTB&=~(1<<MOT2D);
	} else if (motor==2) {
		if (c&MOTA) PORTD|=(1<<MOT3A); else PORTD&=~(1<<MOT3A);
		if (c&MOTB) PORTD|=(1<<MOT3B); else PORTD&=~(1<<MOT3B);
		if (c&MOTC) PORTD|=(1<<MOT3C); else PORTD&=~(1<<MOT3C);
		if (c&MOTD) PORTD|=(1<<MOT3D); else PORTD&=~(1<<MOT3D);
	}
}

//This is where the magic happens... This outputs the actual motor values (motval[0..2]) to the motor.
//The driver uses micro-stepping: the 'real' resolution of the steppers is 8 half-steps per complete
//rotation but we use a resolution of 64 steps per rotation. We get the extra steps by taking the
//64-steps-per-rotation value and adding a counter (pwmpos) running from 0 to 8 to it, then dividing
//the result by 8. This makes the routine do a sort-of-PWM between the two 'real' steps the 
//stepper can have, with the axis of the stepper ending up at the (64-step) angle we set.
//
//What complicates things here is that the steppers can't really run at 10volts. What I do is abuse
//the lower bit of pwmpos to decide if the steppers are powered or not, and take the bits above that as the
//counter value. This makes the software control the motors at a duty cycle of 50%, which seems to make
//them way more happy.
//
//The spit DC motor also is PWMmed, but at a much lower rate. It uses spitpos as the PWM counter.
//
//...and all of this in 16 lines of code.
ISR(TIMER0_COMPA_vect) {
	int x;
	static int pwmpos=0, spitpos=0;
	pwmpos=(pwmpos+1)&7; 

	PORTD&=~(1<<SPIT);
	if (spit && spitpos<35) PORTD|=1<<SPIT;
	spitpos++;
	if (spitpos>200) spitpos=0;

	if (pwmpos&1) {
		disableMotors();
	} else {
		for (x=0; x<3; x++) {
			setMotor(x, (motval[x]+(pwmpos>>1))>>3);
		}
	}
}

#define FPPOS 3
#define ABS(x) (((x)>0)?(x):(-(x)))

//This is a routine that moves the needle point from the current x,y,z-position to a given point.
void moveTo(int x, int y, int z, char doSpit) {
	int steps, i;
	//dx, dy, dz are the amount every motor needs to rotate. Integers.
	int dx=x-motval[0];
	int dy=y-motval[1];
	int dz=z-motval[2];
	//sx,sy,sz are the current coordinates of the needle, in fixed-point notation.
	int sx=motval[0]<<FPPOS;
	int sy=motval[1]<<FPPOS;
	int sz=motval[2]<<FPPOS;
	//The routine will take the maximum amount of steps any direction has and use that to run the main loop
	steps=ABS(dx);
	if (ABS(dy)>steps) steps=ABS(dy);
	if (ABS(dz)>steps) steps=ABS(dz);
	if (steps==0) return;
	//Convert dx,dy,dz into fixed-point values: the amount that motor has to move per iteration of the loop.
	dx=dx/(steps>>FPPOS);
	dy=dy/(steps>>FPPOS);
	dz=dz/(steps>>FPPOS);
	disable=0;
	spit=doSpit;
	//Move!
	for (i=0; i<steps; i++) {
		_delay_us(1200); //Wait a bit for the motors to settle
		sx+=dx; sy+=dy; sz+=dz; //Adjust current (fixed-point) position
		//Tell the interrupt routine that handles the motors the (integer) position we need to be in.
		motval[0]=sx>>FPPOS;
		motval[1]=sy>>FPPOS;
		motval[2]=sz>>FPPOS;
	}
	disable=1; //Movement done!
}


//Structure indicating a single point in the figure array
typedef struct {
	int x;
	int y;
	int z;
	char flags; //see the FL_* macros below
} Point;

#define FL_PEN 1		//If set, pen needs to be down, aka we need to press on the plunger of the syringe.
#define FL_END 2		//Indicates last point of a figure.
#define FL_ENDST 4		//Indicates last point of the figure array.

//Points are given in a random coordinate system that made it easy to define the figures in. This
//macro converts them to actual coordinates the firmware can use.
#define C(x,y,z,f) {x*2,y*2,(z-1800)*3+1300,f}
const Point figs[] PROGMEM={
	C(-250,250,2300, 0),
	C(250,250,2300, FL_PEN),
	C(250,-250,2300, FL_PEN),
	C(-250,-250,2300, FL_PEN),
	C(-250,250,2300, FL_PEN),
	C(-250,250,1800, FL_PEN),
	C(250,250,1800, FL_PEN),
	C(250,250,2300, 0),
	C(250,250,1800, FL_PEN),
	C(250,-250,1800, FL_PEN),
	C(250,-250,2300, 0),
	C(250,-250,1800, FL_PEN),
	C(-250,-250,1800, FL_PEN),
	C(-250,-250,2300, 0),
	C(-250,-250,1800, FL_PEN),
	C(-250,250,1800, FL_PEN|FL_END),
	C( 0, 300, 2500, 0 ),
	C( 110, 278, 2470, FL_PEN ),
	C( 205, 218, 2440, FL_PEN ),
	C( 271, 127, 2410, FL_PEN ),
	C( 299, 18, 2380, FL_PEN ),
	C( 285, -92, 2350, FL_PEN ),
	C( 231, -191, 2320, FL_PEN ),
	C( 144, -262, 2290, FL_PEN ),
	C( 37, -297, 2260, FL_PEN ),
	C( -74, -290, 2230, FL_PEN ),
	C( -176, -242, 2200, FL_PEN ),
	C( -253, -160, 2170, FL_PEN ),
	C( -294, -56, 2140, FL_PEN ),
	C( -294, 56, 2110, FL_PEN ),
	C( -253, 160, 2080, FL_PEN ),
	C( -176, 242, 2050, FL_PEN ),
	C( -74, 290, 2019, FL_PEN ),
	C( 37, 297, 1989, FL_PEN ),
	C( 144, 262, 1959, FL_PEN ),
	C( 231, 191, 1929, FL_PEN ),
	C( 285, 92, 1899, FL_PEN ),
	C( 299, -18, 1869, FL_PEN ),
	C( 271, -127, 1839, FL_PEN ),
	C( 205, -218, 1809, FL_PEN ),
	C( 110, -278, 1779, FL_PEN ),
	C( 0, -299, 1749, FL_PEN ),
	C( -110, -278, 1719, FL_PEN ),
	C( -205, -218, 1689, FL_PEN ),
	C( -271, -127, 1659, FL_PEN ),
	C( -299, -18, 1629, FL_PEN ),
	C( -285, 92, 1599, FL_PEN ),
	C( -231, 191, 1569, FL_PEN ),
	C( -144, 262, 1539, FL_PEN ),
	C( -37, 297, 1509, FL_PEN|FL_END|FL_ENDST ),
};
#undef C

//Main routine
int main(void) {
	int x;
	Point p;
	//Set ports.
	DDRB=0xff;
	DDRD=0x7C;
	PORTD=(1<<BTN);
	//Enable timer
	TCCR0A=0x2;
	TCCR0B=3;
	OCR0A=0x40;
	TIMSK=1;
	sei();
	disable=0;
	//Move the motors to the home position by banging them against the limits...
	for (x=0; x<4000; x++) {
		_delay_us(200);
		motval[1]--;
		motval[2]--;
		motval[0]--;
	}
	motval[0]=0; motval[1]=0; motval[2]=0;
	disable=1;
	//Move to the starting position.
	moveTo(750, 1050, 0, 0);
	motval[0]=0; motval[1]=0; motval[2]=0;

	x=0;
	while(1) {
		while (PIND&(1<<BTN)) ; //Wait for button press
		//Draw a figure: feed points into moveTo until FL_END is seen.
		do {
			memcpy_P(&p, &figs[x], sizeof(Point));
			moveTo(p.x, p.y, p.z, p.flags&FL_PEN);
			x++;
		} while (!(p.flags&FL_END));
		if (p.flags&FL_ENDST) x=0; //Go to the start of the array if the end is reached.
		moveTo(0, 0, 0, 0); //Go to the home position
	}
}
